searchNodes=[{"doc":"Entrypoint module for Ecto related function with Loupe. Ideally, this module should be completely decoupled from any Repo logic and leave that to the app's Repo","ref":"Loupe.Ecto.html","title":"Loupe.Ecto","type":"module"},{"doc":"Same as build_query/2 but with context or with implementation with no assigns","ref":"Loupe.Ecto.html#build_query/2","title":"Loupe.Ecto.build_query/2","type":"function"},{"doc":"Builds an Ecto query from either an AST or a string. It requires an implementation of the Loupe.Ecto.Definition behaviour and supports assigns as a third parameter.","ref":"Loupe.Ecto.html#build_query/4","title":"Loupe.Ecto.build_query/4","type":"function"},{"doc":"","ref":"Loupe.Ecto.html#t:build_query_error/0","title":"Loupe.Ecto.build_query_error/0","type":"type"},{"doc":"The context is the structure that goes through the query building process. It includes the user assigns which are passed down below in the ecto definition, but also includes a couple fields to make sure that query info are validated beforehand. Ecto schema validations Some things are validated before executing the query to avoid crashes during the query execution. These validations includs: Validate that the fetched schema exists and a valid schema. Validate that the queried field are valid on the attached schema and allowed. It also extract the query bindings which are the queries associations and validates them. This is meant to allow someone to use posts.comments.user.name in a query and directly join the posts's comments and comments's user to filter its name. Binding naming Ecto requires named bindings to be atom (which makes total sense). To avoid generating atoms at runtimes in the system, any non-reserverd keyword is cast as string when generating the AST. When building the Ecto query, the join statements uses bindings from a predefined set that you can find a the bottom of this file under the @binding_keys module attribute.","ref":"Loupe.Ecto.Context.html","title":"Loupe.Ecto.Context","type":"module"},{"doc":"Casts a sigil expression using the context's implementation","ref":"Loupe.Ecto.Context.html#cast_sigil/2","title":"Loupe.Ecto.Context.cast_sigil/2","type":"function"},{"doc":"Initializes a query by scoping from the implementation","ref":"Loupe.Ecto.Context.html#initialize_query/1","title":"Loupe.Ecto.Context.initialize_query/1","type":"function"},{"doc":"Instanciates a context with an implementation and some assigns. The assigns will be passed down to the implementation during to query building process to alter the definition dynamically.","ref":"Loupe.Ecto.Context.html#new/3","title":"Loupe.Ecto.Context.new/3","type":"function"},{"doc":"Puts bindings in the context validation that bindings are either valid fields or associations.","ref":"Loupe.Ecto.Context.html#put_bindings/2","title":"Loupe.Ecto.Context.put_bindings/2","type":"function"},{"doc":"Puts parameters compiling sigils and assigning variables","ref":"Loupe.Ecto.Context.html#put_parameters/2","title":"Loupe.Ecto.Context.put_parameters/2","type":"function"},{"doc":"Puts root schema in the context validating it's a valid Ecto schema","ref":"Loupe.Ecto.Context.html#put_root_schema/2","title":"Loupe.Ecto.Context.put_root_schema/2","type":"function"},{"doc":"Checks if a given schema field is allowed","ref":"Loupe.Ecto.Context.html#schema_field_allowed?/3","title":"Loupe.Ecto.Context.schema_field_allowed?/3","type":"function"},{"doc":"Gets implementation schema fields","ref":"Loupe.Ecto.Context.html#schema_fields/2","title":"Loupe.Ecto.Context.schema_fields/2","type":"function"},{"doc":"Gets implementation schemas","ref":"Loupe.Ecto.Context.html#schemas/1","title":"Loupe.Ecto.Context.schemas/1","type":"function"},{"doc":"Same as selectable_fields/1 but for the root schema","ref":"Loupe.Ecto.Context.html#selectable_fields/1","title":"Loupe.Ecto.Context.selectable_fields/1","type":"function"},{"doc":"Gets fields that can be selected infering the foreign keys. For instance, if someone allows user belong_to relationship on Post̀ , the user_id is automatically returned and there is no need for it to be allowed.","ref":"Loupe.Ecto.Context.html#selectable_fields/2","title":"Loupe.Ecto.Context.selectable_fields/2","type":"function"},{"doc":"Sorts binding in order to be join in order","ref":"Loupe.Ecto.Context.html#sorted_bindings/1","title":"Loupe.Ecto.Context.sorted_bindings/1","type":"function"},{"doc":"","ref":"Loupe.Ecto.Context.html#t:assigns/0","title":"Loupe.Ecto.Context.assigns/0","type":"type"},{"doc":"","ref":"Loupe.Ecto.Context.html#t:binding_path/0","title":"Loupe.Ecto.Context.binding_path/0","type":"type"},{"doc":"","ref":"Loupe.Ecto.Context.html#t:bindings/0","title":"Loupe.Ecto.Context.bindings/0","type":"type"},{"doc":"","ref":"Loupe.Ecto.Context.html#t:implementation/0","title":"Loupe.Ecto.Context.implementation/0","type":"type"},{"doc":"","ref":"Loupe.Ecto.Context.html#t:schema/0","title":"Loupe.Ecto.Context.schema/0","type":"type"},{"doc":"","ref":"Loupe.Ecto.Context.html#t:schemas/0","title":"Loupe.Ecto.Context.schemas/0","type":"type"},{"doc":"","ref":"Loupe.Ecto.Context.html#t:t/0","title":"Loupe.Ecto.Context.t/0","type":"type"},{"doc":"Behaviour for defining an Ecto definition. This behaviours defines what the query builder can use to build the query. All function receives the context assigns, meaning that you could add a user's role to the assigns and filter down available schemas for this particular role (example, allow Users query only for Admins) Exmaple In the following example, the Ecto definition prevents non-admin to query User and only allow the :name field for the Game schema. Meaning that non-admin cannot query for other fields than :name when querying games. defmodule MyApp.Loupe.EctoDefinition do @behaviour Loupe.Ecto.Definition @impl Loupe.Ecto.Definition def schemas ( %{ role : &quot;admin&quot; } ) , do : %{ &quot;User&quot; =&gt; MyApp.User , &quot;Game&quot; =&gt; MyApp.Game } def schemas ( _ ) , do : %{ &quot;Game&quot; =&gt; MyApp.Game } @impl Loupe.Ecto.Definition def schema_fields ( Game , %{ role : role } ) when role != &quot;admin&quot; , do : [ :name ] def schema_fields ( _ , _ ) , do : :all @impl Loupe.Ecto.Definition def scope_schemas ( Game , %{ role : role } ) when role != &quot;admin&quot; do Game . not_deleted ( ) end def scope_schemas ( schema , _ ) , do : schema end","ref":"Loupe.Ecto.Definition.html","title":"Loupe.Ecto.Definition","type":"behaviour"},{"doc":"Casts a sigil to another literal","ref":"Loupe.Ecto.Definition.html#c:cast_sigil/3","title":"Loupe.Ecto.Definition.cast_sigil/3","type":"callback"},{"doc":"Gets available field for a schema","ref":"Loupe.Ecto.Definition.html#c:schema_fields/2","title":"Loupe.Ecto.Definition.schema_fields/2","type":"callback"},{"doc":"Gets available schemas","ref":"Loupe.Ecto.Definition.html#c:schemas/1","title":"Loupe.Ecto.Definition.schemas/1","type":"callback"},{"doc":"Scopes schema query builder","ref":"Loupe.Ecto.Definition.html#c:scope_schema/2","title":"Loupe.Ecto.Definition.scope_schema/2","type":"callback"},{"doc":"Module to implement various filter depending on field type. This is how we build query with Ecto depending on the kind of field we receive","ref":"Loupe.Ecto.Filter.html","title":"Loupe.Ecto.Filter","type":"behaviour"},{"doc":"","ref":"Loupe.Ecto.Filter.html#c:apply_bounded_filter/2","title":"Loupe.Ecto.Filter.apply_bounded_filter/2","type":"callback"},{"doc":"Access a composite field dynamically","ref":"Loupe.Ecto.Filter.html#composite_access/2","title":"Loupe.Ecto.Filter.composite_access/2","type":"macro"},{"doc":"Unwraps literal","ref":"Loupe.Ecto.Filter.html#unwrap/2","title":"Loupe.Ecto.Filter.unwrap/2","type":"function"},{"doc":"Defines functions for querying using a composite field from a variant. This is invoked when you refer a field variant using : like: get User where posts . price :amount &gt; 100","ref":"Loupe.Ecto.Filter.CompositeVariant.html","title":"Loupe.Ecto.Filter.CompositeVariant","type":"module"},{"doc":"Defines functions for querying using a direct field. This is invoked when you refer a field directly (without path or variant): get User where posts . title = &quot;Why the sun exists?&quot;","ref":"Loupe.Ecto.Filter.Direct.html","title":"Loupe.Ecto.Filter.Direct","type":"module"},{"doc":"Defines functions for querying using a field with a path. This is invoked when you refer a field plus a path for json fields: get Post where user . role . permissions [ &quot;categories&quot; , &quot;access&quot; ] = &quot;write&quot;","ref":"Loupe.Ecto.Filter.JsonPath.html","title":"Loupe.Ecto.Filter.JsonPath","type":"module"},{"doc":"Loupe's compiler entrypoint.","ref":"Loupe.Language.html","title":"Loupe.Language","type":"module"},{"doc":"Compiles a query to AST","ref":"Loupe.Language.html#compile/1","title":"Loupe.Language.compile/1","type":"function"},{"doc":"","ref":"Loupe.Language.html#t:compile_error/0","title":"Loupe.Language.compile_error/0","type":"type"},{"doc":"","ref":"Loupe.Language.html#t:sigil_definition/0","title":"Loupe.Language.sigil_definition/0","type":"type"},{"doc":"Extracted AST structure query. It uses a basic syntax like [ action ] [ quantifier? ] [ schema ] where [ predicates ] The action is any alphanumeric value. It can be used to specify what do you aim to use the query for. It could ̀ ecto for instance that you query Ecto with, or even ets that lookup an ets table with match spec. The quantifier is used to limit the queries result but can be ommited defaulting to 1 . It supports the following: Positive integer; 1 , 2 , 10 etc... Range: 10..20 , it limits the query to 10 records offsetting to the 10th record all : Returns all the record matching The schema needs to be an idenfifier (non-quoted alphanumeric) that matches the definition's schemas/1 function. The predicates are combination of boolean operators and operand for validation. See the module's type for every support operators but it can basically be a syntax like # could match `get` and run query on Ecto. get 5 User where ( name = &quot;John Doe&quot; ) or ( age &gt; 18 ) # count match `protobuf` to query Ecto and generate Protobufs. protobuf all BoardGame where name like &quot;Catan&quot;","ref":"Loupe.Language.Ast.html","title":"Loupe.Language.Ast","type":"module"},{"doc":"Extracts bindings of an AST","ref":"Loupe.Language.Ast.html#bindings/1","title":"Loupe.Language.Ast.bindings/1","type":"function"},{"doc":"Instanciates the AST","ref":"Loupe.Language.Ast.html#new/5","title":"Loupe.Language.Ast.new/5","type":"function"},{"doc":"Unwraps literal","ref":"Loupe.Language.Ast.html#unwrap_literal/2","title":"Loupe.Language.Ast.unwrap_literal/2","type":"function"},{"doc":"Alpha identifier","ref":"Loupe.Language.Ast.html#t:alpha_identifier/0","title":"Loupe.Language.Ast.alpha_identifier/0","type":"type"},{"doc":"Composed bidings from nested querying","ref":"Loupe.Language.Ast.html#t:binding/0","title":"Loupe.Language.Ast.binding/0","type":"type"},{"doc":"Valid boolean operators","ref":"Loupe.Language.Ast.html#t:boolean_operator/0","title":"Loupe.Language.Ast.boolean_operator/0","type":"type"},{"doc":"","ref":"Loupe.Language.Ast.html#t:external_identifiers/0","title":"Loupe.Language.Ast.external_identifiers/0","type":"type"},{"doc":"Literial values usable in comparison","ref":"Loupe.Language.Ast.html#t:literal/0","title":"Loupe.Language.Ast.literal/0","type":"type"},{"doc":"Negated operator","ref":"Loupe.Language.Ast.html#t:negated_operator/0","title":"Loupe.Language.Ast.negated_operator/0","type":"type"},{"doc":"Json like object decoded","ref":"Loupe.Language.Ast.html#t:object/0","title":"Loupe.Language.Ast.object/0","type":"type"},{"doc":"Valid comparison operands","ref":"Loupe.Language.Ast.html#t:operand/0","title":"Loupe.Language.Ast.operand/0","type":"type"},{"doc":"Operator structure","ref":"Loupe.Language.Ast.html#t:operator/0","title":"Loupe.Language.Ast.operator/0","type":"type"},{"doc":"Parameters provided to the query","ref":"Loupe.Language.Ast.html#t:parameters/0","title":"Loupe.Language.Ast.parameters/0","type":"type"},{"doc":"Validation composed predicates","ref":"Loupe.Language.Ast.html#t:predicate/0","title":"Loupe.Language.Ast.predicate/0","type":"type"},{"doc":"Query quantifier to limit the query result count","ref":"Loupe.Language.Ast.html#t:quantifier/0","title":"Loupe.Language.Ast.quantifier/0","type":"type"},{"doc":"Range from one value to another","ref":"Loupe.Language.Ast.html#t:range/0","title":"Loupe.Language.Ast.range/0","type":"type"},{"doc":"Reserved keywords","ref":"Loupe.Language.Ast.html#t:reserved_keyword/0","title":"Loupe.Language.Ast.reserved_keyword/0","type":"type"},{"doc":"","ref":"Loupe.Language.Ast.html#t:t/0","title":"Loupe.Language.Ast.t/0","type":"type"},{"doc":"Live view component that renders the entities. It also support nesting preloading of the relations though events.","ref":"Loupe.PhoenixLiveView.LiveComponent.html","title":"Loupe.PhoenixLiveView.LiveComponent","type":"module"},{"doc":"Applies a function on an entity at a given path. The path can contain integer, for list indexes or atom for struct's key.","ref":"Loupe.PhoenixLiveView.LiveComponent.html#map_entity_at/3","title":"Loupe.PhoenixLiveView.LiveComponent.map_entity_at/3","type":"function"},{"doc":"Renders a list of struct or a struct recursively. It accepts the following attributes: result (required): either list(struct()) or struct() . definition (required): Ecto definition module. definition_assigns (optional): Assigns for the Ecto definition. inspect_modules (optional): List of module that should be rendered directly using inspect/1 . renderer (optional): Implementation of Loupe.PhoenixLiveView.Renderer behaviour.","ref":"Loupe.PhoenixLiveView.LiveComponent.html#render/1","title":"Loupe.PhoenixLiveView.LiveComponent.render/1","type":"function"},{"doc":"","ref":"Loupe.PhoenixLiveView.LiveComponent.html#t:tree/0","title":"Loupe.PhoenixLiveView.LiveComponent.tree/0","type":"type"},{"doc":"","ref":"Loupe.PhoenixLiveView.LiveComponent.html#t:update_function/0","title":"Loupe.PhoenixLiveView.LiveComponent.update_function/0","type":"type"},{"doc":"Module that is meant to be implemented to control how the component renders","ref":"Loupe.PhoenixLiveView.Renderer.html","title":"Loupe.PhoenixLiveView.Renderer","type":"behaviour"},{"doc":"","ref":"Loupe.PhoenixLiveView.Renderer.html#c:render_type/2","title":"Loupe.PhoenixLiveView.Renderer.render_type/2","type":"callback"},{"doc":"Determines how a value should be rendered for given renderer module","ref":"Loupe.PhoenixLiveView.Renderer.html#render_type/3","title":"Loupe.PhoenixLiveView.Renderer.render_type/3","type":"function"},{"doc":"","ref":"Loupe.PhoenixLiveView.Renderer.html#c:struct_link/3","title":"Loupe.PhoenixLiveView.Renderer.struct_link/3","type":"callback"},{"doc":"Gets a structure's link for a given renderer module","ref":"Loupe.PhoenixLiveView.Renderer.html#struct_link/4","title":"Loupe.PhoenixLiveView.Renderer.struct_link/4","type":"function"},{"doc":"","ref":"Loupe.PhoenixLiveView.Renderer.html#t:t/0","title":"Loupe.PhoenixLiveView.Renderer.t/0","type":"type"},{"doc":"Filters a stream with Loupe's ast. Schema will likely have no impact on the way this filtering is being done","ref":"Loupe.Stream.html","title":"Loupe.Stream","type":"module"},{"doc":"Queries a stream using an AST or a Query. The returned result will be an enumerable function to be executed with Enum functions. For stream queries, the Schema is unused, up to you to perform pre-filtering, put a static one in place or omit it. Examples The following example will filter records whose age is greater than 18 iex&gt; { :ok , stream } = Loupe.Stream . query ( ~s|get where age &gt; 18| , [ ...&gt; %{ age : 76 } , ...&gt; %{ age : 13 } , ...&gt; %{ age : 28 } , ...&gt; %{ age : 6 } , ...&gt; %{ age : 34 } , ...&gt; ] ) iex&gt; Enum . to_list ( stream ) [ %{ age : 76 } ] The same parsing rules applies here, so only one record is returned because when quantifier is provided, it defaults to 1. One could use all to get all the records that are matching or a range. iex&gt; { :ok , stream } = Loupe.Stream . query ( ~s|get all where age &gt; 18| , [ ...&gt; %{ age : 76 } , ...&gt; %{ age : 13 } , ...&gt; %{ age : 28 } , ...&gt; %{ age : 6 } , ...&gt; %{ age : 34 } , ...&gt; ] ) iex&gt; Enum . to_list ( stream ) [ %{ age : 76 } , %{ age : 28 } , %{ age : 34 } ] iex&gt; { :ok , stream } = Loupe.Stream . query ( ~s|get 2..3 where age &gt; 18| , [ ...&gt; %{ age : 76 } , ...&gt; %{ age : 13 } , ...&gt; %{ age : 28 } , ...&gt; %{ age : 6 } , ...&gt; %{ age : 34 } , ...&gt; ] ) iex&gt; Enum . to_list ( stream ) [ %{ age : 28 } , %{ age : 34 } ] Overriding query's limit In case you wanna enforce a limit of your own to the stream and don't wanna depend on the query's quantifier , you can pass limit?: false to the function iex&gt; { :ok , stream } = Loupe.Stream . query ( ~s|get 1 where age &gt; 18| , [ ...&gt; %{ age : 76 } , ...&gt; %{ age : 13 } , ...&gt; %{ age : 28 } , ...&gt; %{ age : 6 } , ...&gt; %{ age : 34 } , ...&gt; ] , limit? : false ) iex&gt; Enum . to_list ( stream ) [ %{ age : 76 } , %{ age : 28 } , %{ age : 34 } ] Using variables You can provide variables to your query with the variables option. Keys must be string to match what is decoded from the query. iex&gt; { :ok , stream } = Loupe.Stream . query ( ~s|get 1 where age &gt; adult| , [ ...&gt; %{ age : 76 } , ...&gt; %{ age : 13 } , ...&gt; %{ age : 28 } , ...&gt; %{ age : 6 } , ...&gt; %{ age : 34 } , ...&gt; ] , limit? : false , variables : %{ &quot;adult&quot; =&gt; 18 } ) iex&gt; Enum . to_list ( stream ) [ %{ age : 76 } , %{ age : 28 } , %{ age : 34 } ]","ref":"Loupe.Stream.html#query/3","title":"Loupe.Stream.query/3","type":"function"},{"doc":"","ref":"Loupe.Stream.html#t:build_query_error/0","title":"Loupe.Stream.build_query_error/0","type":"type"},{"doc":"","ref":"Loupe.Stream.html#t:option/0","title":"Loupe.Stream.option/0","type":"type"},{"doc":"Behaviour to implement comparator. It could be useful to implement your own comparator to alter how fields are getting compared. Overriding the comparator can allow someone to implement variant casting (exmaple field:upper to uppercase values automatically), alter the comparison logic or implement sigil casting.","ref":"Loupe.Stream.Comparator.html","title":"Loupe.Stream.Comparator","type":"behaviour"},{"doc":"Applies a field variant on a value. This can be used to have expression like value:upper be automatically uppercased.","ref":"Loupe.Stream.Comparator.html#c:apply_variant/2","title":"Loupe.Stream.Comparator.apply_variant/2","type":"callback"},{"doc":"Casts a sigil to kind of value to be compared.","ref":"Loupe.Stream.Comparator.html#c:cast_sigil/2","title":"Loupe.Stream.Comparator.cast_sigil/2","type":"callback"},{"doc":"Compares a stream's value with a literal value","ref":"Loupe.Stream.Comparator.html#c:compare/1","title":"Loupe.Stream.Comparator.compare/1","type":"callback"},{"doc":"Compares predicates inside a given map/structure tree","ref":"Loupe.Stream.Comparator.html#compare/3","title":"Loupe.Stream.Comparator.compare/3","type":"function"},{"doc":"Context to work with Stream filtering","ref":"Loupe.Stream.Context.html","title":"Loupe.Stream.Context","type":"module"},{"doc":"Applies AST options in the context to prevent carrying the ast around","ref":"Loupe.Stream.Context.html#apply_ast/2","title":"Loupe.Stream.Context.apply_ast/2","type":"function"},{"doc":"Creates a new context","ref":"Loupe.Stream.Context.html#new/1","title":"Loupe.Stream.Context.new/1","type":"function"},{"doc":"Puts variable in context","ref":"Loupe.Stream.Context.html#put_variables/2","title":"Loupe.Stream.Context.put_variables/2","type":"function"},{"doc":"","ref":"Loupe.Stream.Context.html#t:comparator/0","title":"Loupe.Stream.Context.comparator/0","type":"type"},{"doc":"","ref":"Loupe.Stream.Context.html#t:option/0","title":"Loupe.Stream.Context.option/0","type":"type"},{"doc":"","ref":"Loupe.Stream.Context.html#t:t/0","title":"Loupe.Stream.Context.t/0","type":"type"},{"doc":"Default comparator that does strict comparison. The default comparator does not implement any kind of variant casting or sigil conversion.","ref":"Loupe.Stream.DefaultComparator.html","title":"Loupe.Stream.DefaultComparator","type":"module"},{"doc":"Mock schemas and Ecto modules for tests. For reasons I don't fully understand, seperating this module into seperate files causes compilation issues. Some module aren't loaded properly at the right time and cause problems. Same for some relations, those you see that uses Module.concat/1 , they need to use this call because at the time of reading this file, they aren't properly compiled.","ref":"Loupe.Test.Ecto.html","title":"Loupe.Test.Ecto","type":"module"},{"doc":"","ref":"Loupe.Test.Ecto.html#create_context/1","title":"Loupe.Test.Ecto.create_context/1","type":"function"},{"doc":"","ref":"Loupe.Test.Ecto.html#load_schemas/1","title":"Loupe.Test.Ecto.load_schemas/1","type":"function"},{"doc":"","ref":"Loupe.Test.Ecto.html#run_query/2","title":"Loupe.Test.Ecto.run_query/2","type":"function"},{"doc":"","ref":"Loupe.Test.Ecto.html#sigil_L/2","title":"Loupe.Test.Ecto.sigil_L/2","type":"function"},{"doc":"","ref":"Loupe.Test.Ecto.html#update_assigns/2","title":"Loupe.Test.Ecto.update_assigns/2","type":"function"},{"doc":"","ref":"Loupe.Test.Ecto.html#with_root_schema/1","title":"Loupe.Test.Ecto.with_root_schema/1","type":"function"},{"doc":"Comment schema","ref":"Loupe.Test.Ecto.Comment.html","title":"Loupe.Test.Ecto.Comment","type":"module"},{"doc":"Example Ecto definition for the modules defined above.","ref":"Loupe.Test.Ecto.Definition.html","title":"Loupe.Test.Ecto.Definition","type":"module"},{"doc":"External key schema","ref":"Loupe.Test.Ecto.ExternalKey.html","title":"Loupe.Test.Ecto.ExternalKey","type":"module"},{"doc":"Post schema","ref":"Loupe.Test.Ecto.Post.html","title":"Loupe.Test.Ecto.Post","type":"module"},{"doc":"Mocked repo","ref":"Loupe.Test.Ecto.Repo.html","title":"Loupe.Test.Ecto.Repo","type":"module"},{"doc":"Callback implementation for Ecto.Repo.aggregate/3 .","ref":"Loupe.Test.Ecto.Repo.html#aggregate/3","title":"Loupe.Test.Ecto.Repo.aggregate/3","type":"function"},{"doc":"Callback implementation for Ecto.Repo.aggregate/4 .","ref":"Loupe.Test.Ecto.Repo.html#aggregate/4","title":"Loupe.Test.Ecto.Repo.aggregate/4","type":"function"},{"doc":"Callback implementation for Ecto.Repo.all/2 .","ref":"Loupe.Test.Ecto.Repo.html#all/2","title":"Loupe.Test.Ecto.Repo.all/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.checked_out?/0 .","ref":"Loupe.Test.Ecto.Repo.html#checked_out?/0","title":"Loupe.Test.Ecto.Repo.checked_out?/0","type":"function"},{"doc":"Callback implementation for Ecto.Repo.checkout/2 .","ref":"Loupe.Test.Ecto.Repo.html#checkout/2","title":"Loupe.Test.Ecto.Repo.checkout/2","type":"function"},{"doc":"","ref":"Loupe.Test.Ecto.Repo.html#child_spec/1","title":"Loupe.Test.Ecto.Repo.child_spec/1","type":"function"},{"doc":"Callback implementation for Ecto.Repo.config/0 .","ref":"Loupe.Test.Ecto.Repo.html#config/0","title":"Loupe.Test.Ecto.Repo.config/0","type":"function"},{"doc":"Callback implementation for Ecto.Repo.default_options/1 .","ref":"Loupe.Test.Ecto.Repo.html#default_options/1","title":"Loupe.Test.Ecto.Repo.default_options/1","type":"function"},{"doc":"Callback implementation for Ecto.Repo.delete/2 .","ref":"Loupe.Test.Ecto.Repo.html#delete/2","title":"Loupe.Test.Ecto.Repo.delete/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.delete!/2 .","ref":"Loupe.Test.Ecto.Repo.html#delete!/2","title":"Loupe.Test.Ecto.Repo.delete!/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.delete_all/2 .","ref":"Loupe.Test.Ecto.Repo.html#delete_all/2","title":"Loupe.Test.Ecto.Repo.delete_all/2","type":"function"},{"doc":"A convenience function for SQL-based repositories that forces all connections in the pool to disconnect within the given interval. See Ecto.Adapters.SQL.disconnect_all/3 for more information.","ref":"Loupe.Test.Ecto.Repo.html#disconnect_all/2","title":"Loupe.Test.Ecto.Repo.disconnect_all/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.exists?/2 .","ref":"Loupe.Test.Ecto.Repo.html#exists?/2","title":"Loupe.Test.Ecto.Repo.exists?/2","type":"function"},{"doc":"A convenience function for SQL-based repositories that executes an EXPLAIN statement or similar depending on the adapter to obtain statistics for the given query. See Ecto.Adapters.SQL.explain/4 for more information.","ref":"Loupe.Test.Ecto.Repo.html#explain/3","title":"Loupe.Test.Ecto.Repo.explain/3","type":"function"},{"doc":"Callback implementation for Ecto.Repo.get/3 .","ref":"Loupe.Test.Ecto.Repo.html#get/3","title":"Loupe.Test.Ecto.Repo.get/3","type":"function"},{"doc":"Callback implementation for Ecto.Repo.get!/3 .","ref":"Loupe.Test.Ecto.Repo.html#get!/3","title":"Loupe.Test.Ecto.Repo.get!/3","type":"function"},{"doc":"Callback implementation for Ecto.Repo.get_by/3 .","ref":"Loupe.Test.Ecto.Repo.html#get_by/3","title":"Loupe.Test.Ecto.Repo.get_by/3","type":"function"},{"doc":"Callback implementation for Ecto.Repo.get_by!/3 .","ref":"Loupe.Test.Ecto.Repo.html#get_by!/3","title":"Loupe.Test.Ecto.Repo.get_by!/3","type":"function"},{"doc":"Callback implementation for Ecto.Repo.get_dynamic_repo/0 .","ref":"Loupe.Test.Ecto.Repo.html#get_dynamic_repo/0","title":"Loupe.Test.Ecto.Repo.get_dynamic_repo/0","type":"function"},{"doc":"Callback implementation for Ecto.Repo.in_transaction?/0 .","ref":"Loupe.Test.Ecto.Repo.html#in_transaction?/0","title":"Loupe.Test.Ecto.Repo.in_transaction?/0","type":"function"},{"doc":"Callback implementation for Ecto.Repo.insert/2 .","ref":"Loupe.Test.Ecto.Repo.html#insert/2","title":"Loupe.Test.Ecto.Repo.insert/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.insert!/2 .","ref":"Loupe.Test.Ecto.Repo.html#insert!/2","title":"Loupe.Test.Ecto.Repo.insert!/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.insert_all/3 .","ref":"Loupe.Test.Ecto.Repo.html#insert_all/3","title":"Loupe.Test.Ecto.Repo.insert_all/3","type":"function"},{"doc":"Callback implementation for Ecto.Repo.insert_or_update/2 .","ref":"Loupe.Test.Ecto.Repo.html#insert_or_update/2","title":"Loupe.Test.Ecto.Repo.insert_or_update/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.insert_or_update!/2 .","ref":"Loupe.Test.Ecto.Repo.html#insert_or_update!/2","title":"Loupe.Test.Ecto.Repo.insert_or_update!/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.load/2 .","ref":"Loupe.Test.Ecto.Repo.html#load/2","title":"Loupe.Test.Ecto.Repo.load/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.one/2 .","ref":"Loupe.Test.Ecto.Repo.html#one/2","title":"Loupe.Test.Ecto.Repo.one/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.one!/2 .","ref":"Loupe.Test.Ecto.Repo.html#one!/2","title":"Loupe.Test.Ecto.Repo.one!/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.preload/3 .","ref":"Loupe.Test.Ecto.Repo.html#preload/3","title":"Loupe.Test.Ecto.Repo.preload/3","type":"function"},{"doc":"Callback implementation for Ecto.Repo.prepare_query/3 .","ref":"Loupe.Test.Ecto.Repo.html#prepare_query/3","title":"Loupe.Test.Ecto.Repo.prepare_query/3","type":"function"},{"doc":"Callback implementation for Ecto.Repo.put_dynamic_repo/1 .","ref":"Loupe.Test.Ecto.Repo.html#put_dynamic_repo/1","title":"Loupe.Test.Ecto.Repo.put_dynamic_repo/1","type":"function"},{"doc":"A convenience function for SQL-based repositories that executes the given query. See Ecto.Adapters.SQL.query/4 for more information.","ref":"Loupe.Test.Ecto.Repo.html#query/3","title":"Loupe.Test.Ecto.Repo.query/3","type":"function"},{"doc":"A convenience function for SQL-based repositories that executes the given query. See Ecto.Adapters.SQL.query!/4 for more information.","ref":"Loupe.Test.Ecto.Repo.html#query!/3","title":"Loupe.Test.Ecto.Repo.query!/3","type":"function"},{"doc":"A convenience function for SQL-based repositories that executes the given multi-result query. See Ecto.Adapters.SQL.query_many/4 for more information.","ref":"Loupe.Test.Ecto.Repo.html#query_many/3","title":"Loupe.Test.Ecto.Repo.query_many/3","type":"function"},{"doc":"A convenience function for SQL-based repositories that executes the given multi-result query. See Ecto.Adapters.SQL.query_many!/4 for more information.","ref":"Loupe.Test.Ecto.Repo.html#query_many!/3","title":"Loupe.Test.Ecto.Repo.query_many!/3","type":"function"},{"doc":"Callback implementation for Ecto.Repo.reload/2 .","ref":"Loupe.Test.Ecto.Repo.html#reload/2","title":"Loupe.Test.Ecto.Repo.reload/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.reload!/2 .","ref":"Loupe.Test.Ecto.Repo.html#reload!/2","title":"Loupe.Test.Ecto.Repo.reload!/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.rollback/1 .","ref":"Loupe.Test.Ecto.Repo.html#rollback/1","title":"Loupe.Test.Ecto.Repo.rollback/1","type":"function"},{"doc":"Callback implementation for Ecto.Repo.start_link/1 .","ref":"Loupe.Test.Ecto.Repo.html#start_link/1","title":"Loupe.Test.Ecto.Repo.start_link/1","type":"function"},{"doc":"Callback implementation for Ecto.Repo.stop/1 .","ref":"Loupe.Test.Ecto.Repo.html#stop/1","title":"Loupe.Test.Ecto.Repo.stop/1","type":"function"},{"doc":"Callback implementation for Ecto.Repo.stream/2 .","ref":"Loupe.Test.Ecto.Repo.html#stream/2","title":"Loupe.Test.Ecto.Repo.stream/2","type":"function"},{"doc":"A convenience function for SQL-based repositories that translates the given query to SQL. See Ecto.Adapters.SQL.to_sql/3 for more information.","ref":"Loupe.Test.Ecto.Repo.html#to_sql/2","title":"Loupe.Test.Ecto.Repo.to_sql/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.transaction/2 .","ref":"Loupe.Test.Ecto.Repo.html#transaction/2","title":"Loupe.Test.Ecto.Repo.transaction/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.update/2 .","ref":"Loupe.Test.Ecto.Repo.html#update/2","title":"Loupe.Test.Ecto.Repo.update/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.update!/2 .","ref":"Loupe.Test.Ecto.Repo.html#update!/2","title":"Loupe.Test.Ecto.Repo.update!/2","type":"function"},{"doc":"Callback implementation for Ecto.Repo.update_all/3 .","ref":"Loupe.Test.Ecto.Repo.html#update_all/3","title":"Loupe.Test.Ecto.Repo.update_all/3","type":"function"},{"doc":"Post schema","ref":"Loupe.Test.Ecto.Role.html","title":"Loupe.Test.Ecto.Role","type":"module"},{"doc":"Post schema","ref":"Loupe.Test.Ecto.User.html","title":"Loupe.Test.Ecto.User","type":"module"},{"doc":"User external key relation","ref":"Loupe.Test.Ecto.UserExternalKey.html","title":"Loupe.Test.Ecto.UserExternalKey","type":"module"},{"doc":"This is the base case module for all tests.","ref":"Loupe.TestCase.html","title":"Loupe.TestCase","type":"module"},{"doc":"","ref":"Loupe.TestCase.html#checkout_repo/1","title":"Loupe.TestCase.checkout_repo/1","type":"function"},{"doc":"","ref":"Loupe.TestCase.html#start_repo/1","title":"Loupe.TestCase.start_repo/1","type":"function"},{"doc":"Error that occurs when to schema is provided but is expected","ref":"Loupe.Ecto.Errors.MissingSchemaError.html","title":"Loupe.Ecto.Errors.MissingSchemaError","type":"exception"},{"doc":"","ref":"Loupe.Ecto.Errors.MissingSchemaError.html#t:t/0","title":"Loupe.Ecto.Errors.MissingSchemaError.t/0","type":"type"},{"doc":"Error being raised where there is an error with operators","ref":"Loupe.Ecto.OperatorError.html","title":"Loupe.Ecto.OperatorError","type":"exception"},{"doc":"An error that occured in the lexer's step","ref":"Loupe.Errors.LexerError.html","title":"Loupe.Errors.LexerError","type":"exception"},{"doc":"","ref":"Loupe.Errors.LexerError.html#t:t/0","title":"Loupe.Errors.LexerError.t/0","type":"type"},{"doc":"An error that occured in the parser's step","ref":"Loupe.Errors.ParserError.html","title":"Loupe.Errors.ParserError","type":"exception"},{"doc":"","ref":"Loupe.Errors.ParserError.html#t:t/0","title":"Loupe.Errors.ParserError.t/0","type":"type"},{"doc":"Loupe is query language for Ecto schema inspection in a safe and configurable manner. You can see this example app to understand how it applies with Ecto.","ref":"readme.html","title":"Loupe","type":"extras"},{"doc":"Until Loupe reaches 1.x.x , it's considered experimental. The syntax will change, APIs will change and structure will too. We'll do our best to respect semantic versioning and avoid big breaking changes but they can happen.","ref":"readme.html#important","title":"Loupe - Important","type":"extras"},{"doc":"Loupe is available in Hex , the package can be installed by adding loupe to your list of dependencies in mix.exs : def deps do [ { :loupe , &quot;~&gt; 0.10.0&quot; } ] end Thhe documentation can be found at https://hexdocs.pm/loupe .","ref":"readme.html#installation","title":"Loupe - Installation","type":"extras"},{"doc":"The basic syntax has the following format get [ quantifier? ] [ schema ] [ parameters? ] where [ predicates ] quantifier is how many records you want. You can provide a positive integer ( 1 , 2 , 3 ...), a range ( 1..10 , 10..20 , 50..100 ) or all . schema can be an alphanumeric indentifier that you registered in the Definition (See Ecto Usage for exmaple). The schema is required only for Ecto usage. parameters is a json inspired map. It takes the format of {key: &quot;value&quot;} . Key is an identifier, but value can be any literal type (another object, string, int, float, boolean, list) predicates needs to be a combinaison or operators and boolean operators. Cool stuff You can use k and m quantifiers for numbers. Writing get all User where money &gt; 100k translates to get all User where money &gt; 100000 . Operators The are a couple of basic operators like &lt; , &gt; , &lt;= , &gt;= , = , != . But also some textual operators: in is used with lists, like age in [18, 21] like is used with strings and automatically wraps in % . You can also use the keyword :empty as a null checker like age :empty . Textual operators and :empty can be prefixed with not to negate the expression: not like , not in , age not :empty . For boolean, the binding can be provided as is and prefixed by not for false. Example where active or where not enabled . Boolean Operators So far, the syntax supprts and and or and use parenthese to scope the expressions. Field variant Recently, support for &quot;field variant&quot; has been added. It's a syntax that allows to &quot;customize&quot; a field. The Ecto implementation uses the variant to query composite fields. Assume you have a composite Postgres field that is Money (like the Money.Ecto.Composite.Type type from the Money lib), you can now do the following to query the amount: get User where bank_account :amount &gt;= 1 k Path binding Loupe now supports &quot;Path binding&quot;, being able to specify a path (like a json path) on a field. This is used by te Ecto implementation to query json field like below: get User where role . permissions [ posts , access ] = &quot;write&quot; # or get User where role . permissions [ &quot;posts&quot; , &quot;access&quot; ] = &quot;write&quot; Variables and external identifiers Query variable The library allows you to provide external data to you query. Any identifier (unquote alphanumerical and underscore values) provided on the right side of an operator will be output as such. Taking for instance the Ecto implementation, it allows you to provide external parameter to the query. A good usecase example could be to automatically provide a user_id based from the authenticated user. So you can use it like: get Posts where author_id = user_id Then when evaluating the query you make sure to provide the user id by doing Loupe.Ecto . build_query ( query , EctoDefinition , %{ } , %{ &quot;user_id&quot; =&gt; current_user . id } ) Note : Variables in query are required . When evaluating if the query uses a variable that is not provided, an error will be raise. Parameters This variables can also be used in parameters. Suppose your implementation supports an order_by parameter, you can use it like get Posts { order_by : { direction : direction , field : field } } Unlike variables, they don't need to be provided, they are simply extract as such and it's up to you to manipulate them the way you want. For the case of the Ecto implementation, however, they do need to be implemented so they can be extracted in the returning context.","ref":"readme.html#syntax","title":"Loupe - Syntax","type":"extras"},{"doc":"Create a Definition module The Definition module is necessary for Loupe to work with your Ecto schema. In this module you define the schemas that are allowed to be queried and the fields that are permitted for querying. All callbacks accepts a last argument called &quot;assigns&quot;. The assigns are provided to you when evaluating the query allowing you to alter the defition. You could, for instance, add a user's role to the assign and use that role to filter out the allowed schemas so that only admins can query Users. defmodule MyApp.Loupe.Definition do @moduledoc &quot;&quot;&quot; Example Ecto definition for the modules defined above. &quot;&quot;&quot; @behaviour Loupe.Ecto.Definition @schemas %{ &quot;Post&quot; =&gt; Post , &quot;User&quot; =&gt; User , &quot;Role&quot; =&gt; Role } @impl Loupe.Ecto.Definition def schemas ( %{ role : &quot;admin&quot; } ) , do : @schemas def schemas ( _ ) , do : Map . take ( @schemas , [ &quot;Post&quot; , &quot;User&quot; ] ) @impl Loupe.Ecto.Definition def schema_fields ( _ , %{ role : &quot;admin&quot; } ) , do : :all def schema_fields ( Post , _ ) , do : { :only , [ :title , :body ] } def schema_fields ( User , _ ) , do : { :only , [ :email , :posts ] } def schema_fields ( _ , _ ) , do : :all @impl Loupe.Ecto.Definition def scope_schema ( schema , _ ) , do : schema end Once you have this definition, you can try some queries { :ok , ast } = Loupe.Language . compile ( ~s|get all User where age &gt; 18| ) { :ok , ecto_query } = Loupe.Ecto . build_query ( ast , MyApp.Loupe.Definition , %{ role : &quot;admin&quot; } ) Repo . all ( ecto_query )","ref":"readme.html#ecto-usage","title":"Loupe - Ecto usage","type":"extras"},{"doc":"Support has been added to filter streams or enumerable. The same features applies and some more extra; You can use a quantifier to limit the stream ( get 3 ... ) You can override the whole comparison logic You can use field variant as &quot;modifier&quot; through a custom Loupe.Stream.Comparator implementation. You can use sigil for more complex comparison Example posts = [ %{ title : &quot;My post&quot; , comments : [ %{ comment : &quot;Boring!&quot; , author : &quot;Homer Simpsons&quot; } ] } , %{ title : &quot;My second post&quot; , comments : [ %{ comment : &quot;Respect my authorita!&quot; , author : &quot;Eric Cartman&quot; } ] } , ] { :ok , stream } = Loupe.Stream . query ( ~s|get where comments.author like &quot;Eric&quot;| , posts ) [ %{ title : &quot;My second posts&quot; } ] = Enum . to_list ( stream )","ref":"readme.html#stream-enumerable","title":"Loupe - Stream / Enumerable","type":"extras"},{"doc":"Here are some things that I would like Loupe to support: ~Sorting a query, current ideas involves~ This can be achieve with a parameter like get User{order_by: &quot;age&quot;} where ... and be handled manually by your application ~Support some more complex fields prefixed by ~ (or whatever syntax, inspired by elixir's sigils) like the examples below~ This has been implemented. Field variants can be used for composite fields and sigil can be used for expresions. Implement a LiveView UI lib that shows the strucutres as expandable. Being able to click on a User's posts to automatically preload all of its nested Posts. Also have &quot;block&quot; UI module where you can simply create a query from dropdowns in a form for non-power user. Make lexer and parser swappable. Right now, you are stuck with the internal structure that I came up with. The idea would be to allow some to swap the syntax for anything they want. For instance, a french team could implement a french query language to give to their normal user.","ref":"readme.html#todo","title":"Loupe - Todo","type":"extras"},{"doc":"You can see the CONTRIBUTING.md file to know more about the contributing guidelines. Pull requests are welcome!","ref":"readme.html#contributing","title":"Loupe - Contributing","type":"extras"}]